import typing as t
import re
from ..registry import Header
from ..jwk import KeyFlexible, guess_key
from ..jws import (
    CompactSignature,
    JWSRegistry as _JWSRegistry,
    serialize_compact as _serialize_compact,
    deserialize_compact as _deserialize_compact,
)
from ..util import (
    to_bytes,
    to_str,
    json_b64encode,
    urlsafe_b64encode,
    urlsafe_b64decode,
)
from ..rfc7515.compact import decode_header
from ..errors import BadSignatureError
from .registry import JWSRegistry


def serialize_compact(
        protected: Header,
        payload: t.Union[bytes, str],
        private_key: KeyFlexible,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[_JWSRegistry] = None) -> str:

    if "b64" not in protected:
        return _serialize_compact(protected, payload, private_key, algorithms, registry)

    if registry is None:
        registry = JWSRegistry(algorithms=algorithms)

    if protected["b64"] is True:
        return _serialize_compact(protected, payload, private_key, registry=registry)

    registry.check_header(protected)
    obj = CompactSignature(protected, to_bytes(payload))
    alg = registry.get_alg(protected["alg"])
    key = guess_key(private_key, obj, True)
    key.check_use("sig")

    header_segment = json_b64encode(protected)
    signing_input = header_segment + b"." + obj.payload
    signature = urlsafe_b64encode(alg.sign(signing_input, key))

    # if need to detach payload
    if __is_urlsafe_characters(payload):
        out = signing_input + b"." + signature
    else:
        out = header_segment + b".." + signature
    return out.decode("utf-8")


def deserialize_compact(
        value: t.Union[bytes, str],
        public_key: KeyFlexible,
        payload: t.Optional[t.Union[bytes, str]] = None,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWSRegistry] = None) -> CompactSignature:
    obj = _extract_compact(to_bytes(value), payload)
    if obj is None:
        return _deserialize_compact(value, public_key, algorithms, registry)

    if registry is None:
        registry = JWSRegistry(algorithms=algorithms)

    if obj is True:
        return _deserialize_compact(value, public_key, registry=registry)

    headers = obj.headers()
    registry.check_header(headers)
    key = guess_key(public_key, obj)
    key.check_use("sig")
    alg = registry.get_alg(headers["alg"])

    signing_input = obj.segments["header"] + b"." + obj.payload
    sig = urlsafe_b64decode(obj.segments["signature"])
    if not alg.verify(signing_input, sig, key):
        raise BadSignatureError()
    assert isinstance(obj, CompactSignature)
    return obj


# https://datatracker.ietf.org/doc/html/rfc7797#section-5.2
# the application MUST ensure that the payload contains only the URL-safe
# characters 'a'-'z', 'A'-'Z', '0'-'9', dash ('-'), underscore ('_'),
# and tilde ('~')
_re_urlsafe = re.compile("^[a-zA-Z0-9-_~]+$")


def __is_urlsafe_characters(s: t.Union[bytes, str]) -> bool:
    return bool(_re_urlsafe.match(to_str(s)))


def _extract_compact(value: bytes, payload: t.Optional[t.Union[bytes, str]] = None) -> t.Any:
    parts = value.split(b".")
    if len(parts) != 3:
        raise ValueError("Invalid JSON Web Signature")

    header_segment, payload_segment, signature_segment = parts
    protected = decode_header(header_segment)

    if "b64" not in protected:
        return None

    if protected["b64"] is True:
        return True

    if payload:
        obj = CompactSignature(protected, to_bytes(payload))
    else:
        obj = CompactSignature(protected, payload_segment)
    obj.segments.update({
        "header": header_segment,
        "payload": payload_segment,
        "signature": signature_segment,
    })
    return obj



import json
import typing as t
from .rfc7519.claims import Claims, convert_claims, check_sensitive_data
from .rfc7519.registry import ClaimsOption, JWTClaimsRegistry
from .jws import (
    JWSRegistry,
    serialize_compact,
    deserialize_compact,
)
from .jwe import (
    JWERegistry,
    encrypt_compact,
    decrypt_compact,
)
from .jwk import KeyFlexible
from .errors import InvalidTypeError, InvalidPayloadError
from .util import to_bytes
from .registry import Header

__all__ = [
    "Claims",
    "Token",
    "ClaimsOption",
    "JWTClaimsRegistry",
    "encode",
    "decode",
    "check_sensitive_data",
]

JWTRegistry = t.Union[JWSRegistry, JWERegistry]


class Token:
    """The extracted token object, which contains ``header`` and ``claims``.

    :param header: the header part of the JWT
    :param claims: the payload part of the JWT
    """
    def __init__(self, header: Header, claims: Claims):
        #: header in dict
        self.header = header
        #: payload claims in dict
        self.claims = claims


def encode(
        header: Header,
        claims: Claims,
        key: KeyFlexible,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWTRegistry] = None) -> str:
    """Encode a JSON Web Token with the given header, and claims.

    :param header: A dict of the JWT header
    :param claims: A dict of the JWT claims to be encoded
    :param key: key used to sign the signature
    :param algorithms: a list of allowed algorithms
    :param registry: a ``JWSRegistry`` or ``JWERegistry`` to use
    """
    # add ``typ`` in header
    _header = {"typ": "JWT", **header}
    payload = convert_claims(claims)
    if "enc" in _header:
        if registry is not None:
            assert isinstance(registry, JWERegistry)
        return encrypt_compact(_header, payload, key, algorithms, registry)
    else:
        if registry is not None:
            assert isinstance(registry, JWSRegistry)
        return serialize_compact(_header, payload, key, algorithms, registry)


def decode(
        value: t.Union[bytes, str],
        key: KeyFlexible,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWTRegistry] = None) -> Token:
    """Decode the JSON Web Token string with the given key, and validate
    it with the claims requests.

    :param value: text of the JWT
    :param key: key used to verify the signature
    :param algorithms: a list of allowed algorithms
    :param registry: a ``JWSRegistry`` or ``JWERegistry`` to use
    :raise: BadSignatureError
    """
    _value = to_bytes(value)
    header: Header
    payload: bytes
    if _value.count(b".") == 4:
        if registry is not None:
            assert isinstance(registry, JWERegistry)
        header, payload = _decode_jwe(_value, key, algorithms, registry)
    else:
        if registry is not None:
            assert isinstance(registry, JWSRegistry)
        header, payload = _decode_jws(_value, key, algorithms, registry)

    try:
        claims: Claims = json.loads(payload)
    except (TypeError, ValueError):
        raise InvalidPayloadError()

    token = Token(header, claims)
    typ = token.header.get("typ")
    # https://www.rfc-editor.org/rfc/rfc7519#section-5.1
    # If present, it is RECOMMENDED that its value be "JWT".
    if typ and typ != "JWT":
        raise InvalidTypeError()
    return token


def _decode_jwe(
        value: bytes,
        key: KeyFlexible,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWERegistry] = None) -> t.Tuple[Header, bytes]:
    jwe_obj = decrypt_compact(value, key, algorithms, registry)
    assert jwe_obj.plaintext is not None
    return jwe_obj.headers(), jwe_obj.plaintext


def _decode_jws(
        value: bytes,
        key: KeyFlexible,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWSRegistry] = None) -> t.Tuple[Header, bytes]:
    jws_obj = deserialize_compact(value, key, algorithms, registry)
    assert jws_obj.payload is not None
    return jws_obj.headers(), jws_obj.payload



from tests.base import TestFixture, load_key
from joserfc.jwk import OctKey
from joserfc import jws


class TestJWSExamples(TestFixture):
    def run_test(self, data):
        if "secret" in data:
            key = OctKey.import_key(data["secret"])
            private_key = key
            public_key = key
        else:
            private_key = load_key(data["private_key"])
            public_key = load_key(data["public_key"])

        protected = data["protected"]
        payload = data["payload"]
        algorithms = [protected["alg"]]

        value1 = jws.serialize_compact(protected, payload, private_key, algorithms=algorithms)
        obj1 = jws.deserialize_compact(value1, public_key, algorithms=algorithms)
        self.assertEqual(obj1.protected, protected)
        if "compact" in data:
            self.assertEqual(value1, data["compact"])

        member = {"protected": protected}
        value2 = jws.serialize_json(member, payload, private_key, algorithms=algorithms)
        obj2 = jws.deserialize_json(value2, public_key, algorithms=algorithms)
        self.assertTrue(obj2.flattened)
        if "flattened_json" in data:
            self.assertEqual(value2, data["flattened_json"])

        value3 = jws.serialize_json([member], payload, private_key, algorithms=algorithms)
        obj3 = jws.deserialize_json(value3, public_key, algorithms=algorithms)
        self.assertFalse(obj3.flattened)
        if "general_json" in data:
            self.assertEqual(value3, data["general_json"])


TestJWSExamples.load_fixture("jws_examples.json")



from unittest import TestCase
from joserfc import jws
from joserfc.jwk import RSAKey, OctKey
from joserfc.registry import HeaderParameter
from joserfc.errors import (
    BadSignatureError,
    UnsupportedKeyUseError,
    UnsupportedKeyAlgorithmError,
    UnsupportedKeyOperationError,
    InvalidKeyTypeError,
)
from joserfc.util import urlsafe_b64encode
from tests.base import load_key


class TestJWSErrors(TestCase):
    def test_without_alg(self):
        # missing alg
        self.assertRaises(
            ValueError,
            jws.serialize_compact,
            {"kid": "123"}, "i", "secret"
        )

    def test_none_alg(self):
        header = {"alg": "none"}
        text = jws.serialize_compact(
            header, "i", "secret",
            algorithms=["none"]
        )
        self.assertRaises(
            BadSignatureError,
            jws.deserialize_compact,
            text, "secret",
            algorithms=["none"]
        )

    def test_header_invalid_type(self):
        # kid should be a string
        header = {"alg": "HS256", "kid": 123}
        self.assertRaises(
            ValueError,
            jws.serialize_compact,
            header, "i", "secret",
        )

        # jwk should be a dict
        header = {"alg": "HS256", "jwk": "dict"}
        self.assertRaises(
            ValueError,
            jws.serialize_compact,
            header, "i", "secret",
        )

        # jku should be a URL
        header = {"alg": "HS256", "jku": "url"}
        self.assertRaises(
            ValueError,
            jws.serialize_compact,
            header, "i", "secret",
        )

        # x5c should be a chain of string
        header = {"alg": "HS256", "x5c": "url"}
        self.assertRaises(
            ValueError,
            jws.serialize_compact,
            header, "i", "secret",
        )
        header = {"alg": "HS256", "x5c": [1, 2]}
        self.assertRaises(
            ValueError,
            jws.serialize_compact,
            header, "i", "secret",
        )

    def test_crit_header(self):
        header = {"alg": "HS256", "crit": ["kid"]}
        # missing kid header
        self.assertRaises(
            ValueError,
            jws.serialize_compact,
            header, "i", "secret",
        )

        header = {"alg": "HS256", "kid": "1", "crit": ["kid"]}
        jws.serialize_compact(header, "i", "secret")

    def test_extra_header(self):
        header = {"alg": "HS256", "extra": "hi"}
        self.assertRaises(
            ValueError,
            jws.serialize_compact,
            header, "i", "secret",
        )
        # bypass extra header
        registry = jws.JWSRegistry(strict_check_header=False)
        jws.serialize_compact(header, "i", "secret", registry=registry)

        # or use a header registry
        extra_header = {"extra": HeaderParameter("Extra header", "str", False)}
        registry = jws.JWSRegistry(header_registry=extra_header)
        jws.serialize_compact(header, "i", "secret", registry=registry)

    def test_rsa_invalid_signature(self):
        key1 = RSAKey.generate_key()
        key2 = RSAKey.generate_key()
        header = {"alg": "RS256"}
        text = jws.serialize_compact(header, "i", key1)
        self.assertRaises(
            BadSignatureError,
            jws.deserialize_compact,
            text, key2
        )

        header = {"alg": "PS256"}
        text = jws.serialize_compact(
            header, "i", key1,
            algorithms=["PS256"]
        )
        self.assertRaises(
            BadSignatureError,
            jws.deserialize_compact,
            text, key2,
            algorithms=["PS256"]
        )

    def test_ec_incorrect_curve(self):
        header = {"alg": "ES256"}
        key = load_key("ec-p512-private.pem")
        self.assertRaises(
            ValueError,
            jws.serialize_compact,
            header, "i", key
        )

    def test_ec_invalid_signature(self):
        header = {"alg": "ES256"}
        key1 = load_key("ec-p256-alice.json")
        key2 = load_key("ec-p256-bob.json")
        text = jws.serialize_compact(header, "i", key1)
        self.assertRaises(
            BadSignatureError,
            jws.deserialize_compact,
            text, key2
        )

        parts = text.split('.')
        bad_text = ".".join(parts[:-1]) + "." + urlsafe_b64encode(b"abc").decode("utf-8")
        self.assertRaises(
            BadSignatureError,
            jws.deserialize_compact,
            bad_text, key1
        )

    def test_okp_bad_signature(self):
        header = {"alg": "EdDSA"}
        key1 = load_key("okp-ed448-private.pem")
        key2 = load_key("okp-ed25519-private.json")
        algorithms = ["EdDSA"]
        value = jws.serialize_json({"protected": header}, "i", key1, algorithms=algorithms)
        self.assertRaises(
            BadSignatureError,
            jws.deserialize_json,
            value, key2, algorithms=algorithms,
        )


class TestJWSWithKeyErrors(TestCase):
    def test_invalid_key_use(self):
        key = OctKey.generate_key(parameters={"use": "enc"})
        header = {"alg": "HS256"}
        self.assertRaises(
            UnsupportedKeyUseError,
            jws.serialize_compact,
            header, "i", key
        )

    def test_invalid_key_alg(self):
        key = OctKey.generate_key(parameters={"alg": "HS512"})
        header = {"alg": "HS256"}
        self.assertRaises(
            UnsupportedKeyAlgorithmError,
            jws.serialize_compact,
            header, "i", key
        )

    def test_invalid_key_ops(self):
        key = OctKey.generate_key(parameters={"key_ops": ["verify"]})
        header = {"alg": "HS256"}
        self.assertRaises(
            UnsupportedKeyOperationError,
            jws.serialize_compact,
            header, "i", key
        )

    def test_invalid_key_type(self):
        key = OctKey.generate_key()
        header = {"alg": "RS256"}
        self.assertRaises(
            InvalidKeyTypeError,
            jws.serialize_compact,
            header, "i", key
        )



import typing as t
import random
from .rfc7517.types import AnyKey, KeyParameters, DictKey
from .rfc7518.oct_key import OctKey
from .rfc7518.rsa_key import RSAKey
from .rfc7518.ec_key import ECKey
from .rfc8037.okp_key import OKPKey
from .util import to_bytes

__all__ = [
    "OctKey",
    "RSAKey",
    "ECKey",
    "OKPKey",
    "Key",
    "KeySet",
    "JWKRegistry",
]

Key = t.Union[OctKey, RSAKey, ECKey, OKPKey]


class JWKRegistry:
    """A registry for JWK to record ``joserfc`` supported key types.
    Normally, you would use explicit key types like ``OctKey``, ``RSAKey``;
    This registry provides a way to dynamically import and generate keys.
    For instance:

    .. code-block:: python

        from joserfc.jwk import JWKRegistry

        # instead of choosing which key type to use yourself,
        # JWKRegistry can import it automatically
        data = {"kty": "oct", "k": "..."}
        key = JWKRegistry.import_key(data)
    """
    key_types: t.Dict[str, t.Type[Key]] = {
        OctKey.key_type: OctKey,
        RSAKey.key_type: RSAKey,
        ECKey.key_type: ECKey,
        OKPKey.key_type: OKPKey,
    }

    @classmethod
    def import_key(
            cls,
            data: AnyKey,
            key_type: t.Optional[str] = None,
            parameters: t.Optional[KeyParameters] = None) -> Key:
        """A class method for importing a key from bytes, string, and dict.
        When ``value`` is a dict, this method can tell the key type automatically,
        otherwise, developers SHOULD pass the ``key_type`` themselves.

        :param data: the key data in bytes, string, or dict.
        :param key_type: an optional key type in string.
        :param parameters: extra key parameters
        :return: OctKey, RSAKey, ECKey, or OKPKey
        """
        if isinstance(data, dict) and key_type is None:
            if "kty" in data:
                key_type = data["kty"]  # type: ignore[assignment]
            else:
                raise ValueError("Missing key type")

        if key_type not in cls.key_types:
            raise ValueError(f'Invalid key type: "{key_type}"')

        if isinstance(data, str):
            data = to_bytes(data)

        key_cls = cls.key_types[key_type]
        return key_cls.import_key(data, parameters)

    @classmethod
    def generate_key(
            cls,
            key_type: str,
            crv_or_size: t.Union[str, int],
            parameters: t.Optional[KeyParameters] = None,
            private: bool = True,
            auto_kid: bool = False) -> Key:
        """A class method for generating key according to the given key type.
        When ``key_type`` is "oct" and "RSA", the second parameter SHOULD be
        a key size in bits. When ``key_type`` is "EC" and "OKP", the second
        parameter SHOULD be a "crv" string.

        .. code-block:: python

            JWKRegistry.generate_key("RSA", 2048)
            JWKRegistry.generate_key("EC", "P-256")
        """
        if key_type not in cls.key_types:
            raise ValueError(f'Invalid key type: "{key_type}"')

        key_cls = cls.key_types[key_type]
        return key_cls.generate_key(crv_or_size, parameters, private, auto_kid)  # type: ignore[arg-type]


KeySetSerialization = t.TypedDict("KeySetSerialization", {"keys": t.List[DictKey]})


class KeySet:
    #: keys in the key set
    keys: t.List[Key]

    registry_cls: t.Type[JWKRegistry] = JWKRegistry
    algorithm_keys: t.ClassVar[t.Dict[str, t.List[str]]] = {}

    def __init__(self, keys: t.List[Key]):
        for key in keys:
            key.ensure_kid()
        self.keys = keys

    def __iter__(self) -> t.Iterator[Key]:
        return iter(self.keys)

    def __bool__(self) -> bool:
        return bool(self.keys)

    def as_dict(self, private: t.Optional[bool] = None, **params: t.Any) -> KeySetSerialization:
        keys: t.List[DictKey] = []

        for key in self.keys:
            # trigger key to generate kid via thumbprint
            key.ensure_kid()
            if isinstance(key, OctKey):
                keys.append(key.as_dict(**params))
            else:
                keys.append(key.as_dict(private=private, **params))
        return {"keys": keys}

    def get_by_kid(self, kid: t.Optional[str] = None) -> Key:
        if kid is None and len(self.keys) == 1:
            return self.keys[0]

        for key in self.keys:
            if key.kid == kid:
                return key
        raise ValueError(f'No key for kid: "{kid}"')

    def pick_random_key(self, algorithm: str) -> t.Optional[Key]:
        key_types = self.algorithm_keys.get(algorithm)
        if key_types:
            keys = [k for k in self.keys if k.key_type in key_types]
        else:
            keys = self.keys
        if keys:
            return random.choice(keys)
        return None

    @classmethod
    def import_key_set(
            cls,
            value: KeySetSerialization,
            parameters: t.Optional[KeyParameters] = None) -> "KeySet":
        keys: t.List[Key] = []

        for data in value["keys"]:
            keys.append(cls.registry_cls.import_key(data, parameters=parameters))

        return cls(keys)

    @classmethod
    def generate_key_set(
            cls,
            key_type: str,
            crv_or_size: t.Union[str, int],
            parameters: t.Optional[KeyParameters] = None,
            private: bool = True,
            count: int = 4) -> "KeySet":

        keys: t.List[Key] = []
        for _ in range(count):
            key = cls.registry_cls.generate_key(key_type, crv_or_size, parameters, private)
            keys.append(key)

        return cls(keys)



from .registry import JWSRegistry
from .compact import serialize_compact, deserialize_compact
from .json import serialize_json, deserialize_json

__all__ = [
    "JWSRegistry",
    "serialize_compact",
    "deserialize_compact",
    "serialize_json",
    "deserialize_json",
]



import typing as t

__all__ = [
    "JSONRecipientDict",
    "FlattenedJSONSerialization",
    "GeneralJSONSerialization",
]


class JSONRecipientDict(t.TypedDict, total=False):
    header: t.Dict[str, t.Any]
    encrypted_key: str


class GeneralJSONSerialization(t.TypedDict, total=False):
    protected: str
    unprotected: t.Dict[str, t.Any]
    iv: str
    aad: str
    ciphertext: str
    tag: str
    recipients: t.List[JSONRecipientDict]


class FlattenedJSONSerialization(t.TypedDict, total=False):
    protected: str
    unprotected: t.Dict[str, t.Any]
    iv: str
    aad: str
    ciphertext: str
    tag: str
    header: t.Dict[str, t.Any]
    encrypted_key: str



from tests.base import TestFixture, load_key
from joserfc import jws
from joserfc.util import urlsafe_b64encode


# https://datatracker.ietf.org/doc/html/rfc7520#section-4
payload = (
    b"It\xe2\x80\x99s a dangerous business, Frodo, going out your door. "
    b"You step onto the road, and if you don't keep your feet, "
    b"there\xe2\x80\x99s no knowing where you might be swept off to."
)
b64_payload = urlsafe_b64encode(payload).decode("utf-8")


class TestJWSRFC7520(TestFixture):
    def run_test(self, data):
        private_key = load_key(data["private_key"])
        public_key = load_key(data["public_key"])

        protected = data["protected"]
        algorithms = [protected["alg"]]

        obj1 = jws.deserialize_compact(data["compact"], public_key, algorithms=algorithms)
        self.assertEqual(obj1.payload, payload)

        obj2 = jws.deserialize_json(data["general_json"], public_key, algorithms=algorithms)
        self.assertEqual(obj2.payload, payload)
        self.assertFalse(obj2.flattened)

        obj3 = jws.deserialize_json(data["flattened_json"], public_key, algorithms=algorithms)
        self.assertEqual(obj3.payload, payload)
        self.assertTrue(obj3.flattened)

        # try serialize and deserialize pair
        value4 = jws.serialize_compact(protected, payload, private_key, algorithms=algorithms)
        obj4 = jws.deserialize_compact(value4, public_key, algorithms=algorithms)
        self.assertEqual(obj4.payload, payload)

        member = {"protected": protected}
        value5 = jws.serialize_json([member], payload, private_key, algorithms=algorithms)
        obj5 = jws.deserialize_json(value5, public_key, algorithms=algorithms)
        self.assertEqual(obj5.payload, payload)
        self.assertFalse(obj5.flattened)

        value6 = jws.serialize_json(member, payload, private_key, algorithms=algorithms)
        obj6 = jws.deserialize_json(value6, public_key, algorithms=algorithms)
        self.assertEqual(obj6.payload, payload)
        self.assertTrue(obj6.flattened)

        # signature won't change with these algorithms
        if protected["alg"].startswith(("HS", "RS")):
            self.assertEqual(value4, data["compact"])
            self.assertEqual(value5, data["general_json"])
            self.assertEqual(value6, data["flattened_json"])

    def test_signature_with_detached_content(self):
        # https://datatracker.ietf.org/doc/html/rfc7520#section-4.5
        protected = {
            "alg": "HS256",
            "kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037"
        }
        key = load_key("RFC7520-oct-sig.json")
        value1 = jws.serialize_compact(protected, payload, key)
        compact_result = (
            "eyJhbGciOiJIUzI1NiIsImtpZCI6IjAxOGMwYWU1LTRkOWItNDcxYi1iZmQ2LW"
            "VlZjMxNGJjNzAzNyJ9"
            ".."
            "s0h6KThzkfBBBkLspW1h84VsJZFTsPPqMDA7g1Md7p0"
        )
        self.assertEqual(jws.detach_content(value1), compact_result)

        member = {"protected": protected}
        value2 = jws.serialize_json([member], payload, key)
        general_json = {
            "signatures": [
                {
                    "protected": (
                        "eyJhbGciOiJIUzI1NiIsImtpZCI6IjAxOGMwYWU1LTRkOWItNDcx"
                        "Yi1iZmQ2LWVlZjMxNGJjNzAzNyJ9"
                    ),
                    "signature": "s0h6KThzkfBBBkLspW1h84VsJZFTsPPqMDA7g1Md7p0"
                }
            ]
        }
        self.assertEqual(jws.detach_content(value2), general_json)

        flattened_json = {
            "protected": (
                "eyJhbGciOiJIUzI1NiIsImtpZCI6IjAxOGMwYWU1LTRkOWItNDcx"
                "Yi1iZmQ2LWVlZjMxNGJjNzAzNyJ9"
            ),
            "signature": "s0h6KThzkfBBBkLspW1h84VsJZFTsPPqMDA7g1Md7p0"
        }
        value3 = jws.serialize_json(member, payload, key)
        self.assertEqual(jws.detach_content(value3), flattened_json)

    def test_protecting_specific_header_fields(self):
        # https://datatracker.ietf.org/doc/html/rfc7520#section-4.6
        protected = {"alg": "HS256"}
        unprotected = {"kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037"}
        member = {"protected": protected, "header": unprotected}
        key = load_key("RFC7520-oct-sig.json")

        value1 = jws.serialize_json([member], payload, key)
        general_json = {
            "payload": b64_payload,
            "signatures": [
                {
                    "protected": "eyJhbGciOiJIUzI1NiJ9",
                    "header": {"kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037"},
                    "signature": "bWUSVaxorn7bEF1djytBd0kHv70Ly5pvbomzMWSOr20"
                }
            ]
        }
        self.assertEqual(value1, general_json)

        value2 = jws.serialize_json(member, payload, key)
        flattened_json = {
            "payload": b64_payload,
            "protected": "eyJhbGciOiJIUzI1NiJ9",
            "header": {"kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037"},
            "signature": "bWUSVaxorn7bEF1djytBd0kHv70Ly5pvbomzMWSOr20"
        }
        self.assertEqual(value2, flattened_json)

    def test_protecting_content_only(self):
        # https://datatracker.ietf.org/doc/html/rfc7520#section-4.7
        unprotected = {
            "alg": "HS256",
            "kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037"
        }
        key = load_key("RFC7520-oct-sig.json")
        member = {"header": unprotected}
        value1 = jws.serialize_json([member], payload, key)
        general_json = {
            "payload": b64_payload,
            "signatures": [
                {
                    "header": {
                        "alg": "HS256",
                        "kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037"
                    },
                    "signature": "xuLifqLGiblpv9zBpuZczWhNj1gARaLV3UxvxhJxZuk"
                }
            ]
        }
        self.assertEqual(value1, general_json)

        value2 = jws.serialize_json(member, payload, key)
        flattened_json = {
            "payload": b64_payload,
            "header": {
                "alg": "HS256",
                "kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037"
            },
            "signature": "xuLifqLGiblpv9zBpuZczWhNj1gARaLV3UxvxhJxZuk"
        }
        self.assertEqual(value2, flattened_json)

    def test_multiple_signatures(self):
        # https://www.rfc-editor.org/rfc/rfc7520#section-4.8
        output = {
            "payload": (
                "SXTigJlzIGEgZGFuZ2Vyb3VzIGJ1c2luZXNzLCBGcm9kbywgZ29pbmcgb"
                "3V0IHlvdXIgZG9vci4gWW91IHN0ZXAgb250byB0aGUgcm9hZCwgYW5kIGl"
                "mIHlvdSBkb24ndCBrZWVwIHlvdXIgZmVldCwgdGhlcmXigJlzIG5vIGtub"
                "3dpbmcgd2hlcmUgeW91IG1pZ2h0IGJlIHN3ZXB0IG9mZiB0by4"
            ),
            "signatures": [
                {
                    "protected": "eyJhbGciOiJSUzI1NiJ9",
                    "header": {
                        "kid": "bilbo.baggins@hobbiton.example"
                    },
                    "signature": (
                        "MIsjqtVlOpa71KE-Mss8_Nq2YH4FGhiocsqrgi5NvyG53uoimic1tc"
                        "MdSg-qptrzZc7CG6Svw2Y13TDIqHzTUrL_lR2ZFcryNFiHkSw129Egh"
                        "GpwkpxaTn_THJTCglNbADko1MZBCdwzJxwqZc-1RlpO2HibUYyXSwO9"
                        "7BSe0_evZKdjvvKSgsIqjytKSeAMbhMBdMma622_BG5t4sdbuCHtFjp"
                        "9iJmkio47AIwqkZV1aIZsv33uPUqBBCXbYoQJwt7mxPftHmNlGoOSMx"
                        "R_3thmXTCm4US-xiNOyhbm8afKK64jU6_TPtQHiJeQJxz9G3Tx-083B745_AfYOnlC9w"
                    )
                },
                {
                    "header": {
                        "alg": "ES512",
                        "kid": "bilbo.baggins@hobbiton.example"
                    },
                    "signature": (
                        "ARcVLnaJJaUWG8fG-8t5BREVAuTY8n8YHjwDO1muhcdCoFZFFjfISu0"
                        "Cdkn9Ybdlmi54ho0x924DUz8sK7ZXkhc7AFM8ObLfTvNCrqcI3Jkl2U"
                        "5IX3utNhODH6v7xgy1Qahsn0fyb4zSAkje8bAWz4vIfj5pCMYxxm4fgV3q7ZYhm5eD"
                    )
                },
                {
                    "protected": "eyJhbGciOiJIUzI1NiIsImtpZCI6IjAxOGMwYWU1LTRkOWItNDcxYi1iZmQ2LWVlZjMxNGJjNzAzNyJ9",
                    "signature": "s0h6KThzkfBBBkLspW1h84VsJZFTsPPqMDA7g1Md7p0"
                }
            ]
        }

        def resolve_key(recipient):
            headers = recipient.headers()
            if headers["alg"] == "ES512":
                return load_key("RFC7520-EC-public.json")
            elif headers["alg"] == "RS256":
                return load_key("RFC7520-RSA-public.json")
            return load_key("RFC7520-oct-sig.json")

        algorithms = ["RS256", "ES512", "HS256"]
        obj = jws.deserialize_json(output, resolve_key, algorithms=algorithms)
        self.assertEqual([m.headers()["alg"] for m in obj.members], algorithms)


TestJWSRFC7520.load_fixture("jws_rfc7520.json")



from typing import Dict, List, Optional
from .model import JWSAlgModel
from ..registry import (
    JWS_HEADER_REGISTRY,
    Header,
    HeaderRegistryDict,
    validate_registry_header,
    check_crit_header,
    check_supported_header,
)


class JWSRegistry:
    """A registry for JSON Web Signature to keep all the supported algorithms.
    An instance of ``JWSRegistry`` is usually used together with methods in
    ``joserfc.jws``.

    :param header_registry: extra header parameters registry
    :param algorithms: allowed algorithms to be used
    :param strict_check_header: only allow header key in the registry to be used
    """
    default_header_registry: HeaderRegistryDict = JWS_HEADER_REGISTRY
    algorithms: Dict[str, JWSAlgModel] = {}
    recommended: List[str] = []

    def __init__(
            self,
            header_registry: Optional[HeaderRegistryDict] = None,
            algorithms: Optional[List[str]] = None,
            strict_check_header: bool = True):
        self.header_registry: HeaderRegistryDict = {}
        self.header_registry.update(self.default_header_registry)
        if header_registry is not None:
            self.header_registry.update(header_registry)
        self.allowed = algorithms
        self.strict_check_header = strict_check_header

    @classmethod
    def register(cls, alg: JWSAlgModel) -> None:
        """Register a given JWS algorithm instance to the registry."""
        cls.algorithms[alg.name] = alg
        if alg.recommended:
            cls.recommended.append(alg.name)

    def get_alg(self, name: str) -> JWSAlgModel:
        """Get the allowed algorithm instance of the given name.

        :param name: value of the ``alg``, e.g. ``HS256``, ``RS256``
        """
        if name not in self.algorithms:
            raise ValueError(f'Algorithm of "{name}" is not supported')
        if self.allowed:
            allowed = self.allowed
        else:
            allowed = self.recommended

        if name not in allowed:
            raise ValueError(f'Algorithm of "{name}" is not allowed')
        return self.algorithms[name]

    def check_header(self, header: Header) -> None:
        """Check and validate the fields in header part of a JWS object."""
        check_crit_header(header)
        validate_registry_header(self.header_registry, header)
        if self.strict_check_header:
            check_supported_header(self.header_registry, header)


#: default JWS registry
default_registry = JWSRegistry()


def construct_registry(algorithms: Optional[List[str]] = None) -> JWSRegistry:
    if algorithms:
        registry = JWSRegistry(algorithms=algorithms)
    else:
        registry = default_registry
    return registry



from unittest import TestCase
from joserfc.jws import JWSRegistry, serialize_compact, deserialize_compact
from joserfc.jwk import OctKey, RSAKey, KeySet
from joserfc.errors import BadSignatureError, DecodeError, MissingAlgorithmError


class TestCompact(TestCase):
    def test_registry_is_none(self):
        value = serialize_compact({"alg": "HS256"}, b"foo", "secret")
        expected = 'eyJhbGciOiJIUzI1NiJ9.Zm9v.0pehoi-RMZM1jl-4TP_C4Y6BJ-bcmsuzfDyQpkpJkh0'
        self.assertEqual(value, expected)

        obj = deserialize_compact(value, "secret")
        self.assertEqual(obj.payload, b"foo")

    def test_bad_signature_error(self):
        value = b'eyJhbGciOiJIUzI1NiJ9.Zm9v.0pehoi-RMZM1jl-4TP_C4Y6BJ-bcmsuzfDyQpkpJkh0'
        self.assertRaises(BadSignatureError, deserialize_compact, value, "incorrect")

    def test_raise_none_supported_alg(self):
        self.assertRaises(ValueError, serialize_compact, {"alg": "HS512"}, b"foo", "secret")
        self.assertRaises(ValueError, serialize_compact, {"alg": "NOT"}, b"foo", "secret")

    def test_invalid_length(self):
        self.assertRaises(ValueError, deserialize_compact, b'a.b.c.d', "secret")

    def test_no_invalid_header(self):
        # invalid base64
        value = b'abc.Zm9v.0pehoi'
        self.assertRaises(DecodeError, deserialize_compact, value, "secret")

        # no alg value
        value = b'eyJhIjoiYiJ9.Zm9v.0pehoi'
        self.assertRaises(MissingAlgorithmError, deserialize_compact, value, "secret")

    def test_invalid_payload(self):
        value = b'eyJhbGciOiJIUzI1NiJ9.a$b.0pehoi'
        self.assertRaises(DecodeError, deserialize_compact, value, "secret")

    def test_with_key_set(self):
        keys = KeySet([
            OctKey.import_key("a"),
            OctKey.import_key("b"),
            OctKey.import_key("c"),
        ])
        value = serialize_compact({"alg": "HS256"}, b"foo", keys)
        obj = deserialize_compact(value, keys)
        self.assertEqual(obj.payload, b"foo")

        keys.keys.append(RSAKey.generate_key(auto_kid=True))
        value = serialize_compact({"alg": "RS256"}, b"foo", keys)
        obj = deserialize_compact(value, keys)
        self.assertEqual(obj.payload, b"foo")

    def test_strict_check_header(self):
        header = {"alg": "HS256", "custom": "hi"}
        self.assertRaises(ValueError, serialize_compact, header, b"hi", "secret")

        registry = JWSRegistry(strict_check_header=False)
        serialize_compact(header, b"hi", "secret", registry=registry)



from typing import Tuple
from Crypto.Cipher import ChaCha20_Poly1305
from ..rfc7516.registry import JWERegistry
from ..rfc7516.models import JWEEncModel

__all__ = ["ChaCha20EncModel", "JWE_ENC_MODELS", "register_chaha20_poly1305"]


class ChaCha20EncModel(JWEEncModel):
    # https://datatracker.ietf.org/doc/html/draft-amringer-jose-chacha-02#section-4
    cek_size = 256
    recommended = False

    def __init__(self, name: str, description: str, iv_size: int):
        self.name = name
        self.description = description
        self.iv_size = iv_size

    def encrypt(self, plaintext: bytes, cek: bytes, iv: bytes, aad: bytes) -> Tuple[bytes, bytes]:
        """Key Encryption with AEAD_CHACHA20_POLY1305
        """
        chacha = ChaCha20_Poly1305.new(key=cek, nonce=iv)
        chacha.update(aad)
        ciphertext, tag = chacha.encrypt_and_digest(plaintext)
        return ciphertext, tag

    def decrypt(self, ciphertext: bytes, tag: bytes, cek: bytes, iv: bytes, aad: bytes) -> bytes:
        """Key Decryption with AEAD_CHACHA20_POLY1305."""
        chacha = ChaCha20_Poly1305.new(key=cek, nonce=iv)
        chacha.update(aad)
        return chacha.decrypt_and_verify(ciphertext, tag)


C20P = ChaCha20EncModel("C20P", "ChaCha20-Poly1305", 96)
XC20P = ChaCha20EncModel("XC20P", "XChaCha20-Poly1305", 192)

JWE_ENC_MODELS = [C20P, XC20P]


def register_chaha20_poly1305() -> None:
    for model in JWE_ENC_MODELS:
        JWERegistry.register(model)



import json
from pathlib import Path

BASE_PATH = Path(__file__).parent


def read_key(filename: str):
    with open((BASE_PATH / filename).resolve(), "rb") as f:
        content: bytes = f.read()

    if filename.endswith(".json"):
        return json.loads(content)
    return content



from unittest import TestCase
from joserfc import util


class TestUtil(TestCase):
    def test_to_bytes(self):
        self.assertEqual(util.to_bytes(b'foo'), b'foo')
        self.assertEqual(util.to_bytes('foo'), b'foo')
        self.assertEqual(util.to_bytes(123), b'123')
        self.assertEqual(util.to_bytes([102, 111, 111]), b'foo')

    def test_to_unicode(self):
        self.assertEqual(util.to_str(b'foo'), 'foo')
        self.assertEqual(util.to_str('foo'), 'foo')

    def test_int_to_base64(self):
        self.assertRaises(
            ValueError,
            util.int_to_base64,
            -1
        )

    def test_json_b64encode(self):
        self.assertEqual(util.json_b64encode("{}"), b"e30")



from joserfc.jwk import OctKey
from joserfc.rfc7797 import (
    JWSRegistry,
    serialize_compact,
    deserialize_compact,
    serialize_json,
    deserialize_json,
)
from joserfc.errors import (
    DecodeError,
    MissingAlgorithmError,
    BadSignatureError,
)
from joserfc.util import to_bytes
from joserfc import jws
from tests.base import TestFixture


key = OctKey.import_key({
    "kty": "oct",
    "k": (
        "AyM1SysPpbyDfgZld3umj1qzKObwVMkoqQ-EstJQLr_T-1qS0gZH75aKtMN3Yj"
        "0iPS4hcgUuTwjAzZr1Z9CAow"
    )
})


class TestRFC7797(TestFixture):
    def run_test(self, data):
        protected = data["protected"]
        payload = data["payload"]
        value1 = serialize_compact(protected, payload, key)
        self.assertEqual(value1, data["compact"])
        obj1 = deserialize_compact(value1, key, payload=payload)
        self.assertEqual(obj1.headers(), protected)
        self.assertEqual(obj1.payload, to_bytes(payload))
        value2 = serialize_json({"protected": protected}, payload, key)
        self.assertEqual(value2, data["flattened_json"])
        obj2 = deserialize_json(value2, key)
        self.assertTrue(obj2.flattened)
        self.assertEqual(obj2.payload, to_bytes(payload))
        self.assertEqual(obj2.members[0].protected, protected)

    def test_b64_without_crit(self):
        protected = {"alg": "HS256", "b64": False}
        self.assertRaises(
            ValueError,
            serialize_compact,
            protected, "i", key
        )

    def test_invalid_b64_value(self):
        protected = {"alg": "HS256", "b64": "true", "crit": ["b64"]}
        self.assertRaises(
            ValueError,
            serialize_compact,
            protected, "i", key
        )

    def test_compact_invalid_value_length(self):
        self.assertRaises(
            ValueError,
            deserialize_compact,
            b"a.b.c.d.e", key
        )

    def test_invalid_header(self):
        self.assertRaises(
            DecodeError,
            deserialize_compact,
            b"a.b.c", key
        )

    def test_compact_missing_alg(self):
        self.assertRaises(
            MissingAlgorithmError,
            deserialize_compact,
            b"e30.a.b", key
        )

    def test_compact_bad_signature(self):
        protected = {"alg": "HS256", "b64": False, "crit": ["b64"]}
        value = serialize_compact(protected, "hello", key)
        self.assertRaises(
            BadSignatureError,
            deserialize_compact,
            value, "secret"
        )

    def test_compact_use_registry(self):
        registry = JWSRegistry()
        protected = {"alg": "HS256", "b64": True, "crit": ["b64"]}
        value = serialize_compact(protected, "hello", key, registry=registry)
        obj = deserialize_compact(value, key, registry=registry)
        self.assertEqual(obj.protected, protected)

        protected = {"alg": "HS256"}
        value = serialize_compact(protected, "hello", key, registry=registry)
        obj = deserialize_compact(value, key, registry=registry)
        self.assertEqual(obj.protected, protected)

    def test_json_without_protected_header(self):
        registry = JWSRegistry()
        header = {"alg": "HS256", "b64": False, "crit": ["b64"]}
        member = {"header": header}
        value = serialize_json(member, "hello", key, registry=registry)
        obj = deserialize_json(value, key, registry=registry)
        self.assertTrue(obj.flattened)
        self.assertEqual(obj.headers(), header)

    def test_general_json(self):
        members = [{"protected": {"alg": "HS256"}}]
        value = jws.serialize_json(members, "hello", key)
        obj = deserialize_json(value, key)
        self.assertFalse(obj.flattened)

    def test_json_bad_signature(self):
        member = {"protected": {"alg": "HS256", "b64": False, "crit": ["b64"]}}
        value = serialize_json(member, "hello", key)
        self.assertRaises(
            BadSignatureError,
            deserialize_json,
            value, "secret"
        )


TestRFC7797.load_fixture("jws_rfc7797.json")



import typing as t
from .models import JWEAlgModel, JWEEncModel, JWEZipModel
from ..registry import (
    Header,
    HeaderRegistryDict,
    JWE_HEADER_REGISTRY,
    check_supported_header,
    validate_registry_header,
    check_crit_header,
)

JWEAlgorithm = t.Union[JWEAlgModel, JWEEncModel, JWEZipModel]

AlgorithmsDict = t.TypedDict("AlgorithmsDict", {
    "alg": t.Dict[str, JWEAlgModel],
    "enc": t.Dict[str, JWEEncModel],
    "zip": t.Dict[str, JWEZipModel],
})


class JWERegistry:
    """A registry for JSON Web Encryption to keep all the supported algorithms.
    An instance of ``JWERegistry`` is usually used together with methods in
    ``joserfc.jwe``.

    :param header_registry: extra header parameters registry
    :param algorithms: allowed algorithms to be used
    :param strict_check_header: only allow header key in the registry to be used
    """
    algorithms: t.ClassVar[AlgorithmsDict] = {
        "alg": {},
        "enc": {},
        "zip": {},
    }
    recommended: t.ClassVar[t.List[str]] = []

    def __init__(
            self,
            header_registry: t.Optional[HeaderRegistryDict] = None,
            algorithms: t.Optional[t.List[str]] = None,
            strict_check_header: bool = True):
        self.header_registry: HeaderRegistryDict = {}
        self.header_registry.update(JWE_HEADER_REGISTRY)
        if header_registry is not None:
            self.header_registry.update(header_registry)
        self.allowed = algorithms
        self.strict_check_header = strict_check_header

    @classmethod
    def register(cls, model: JWEAlgorithm):
        cls.algorithms[model.algorithm_location][model.name] = model  # type: ignore
        if model.recommended:
            cls.recommended.append(model.name)

    def check_header(self, header: Header, check_more=False):
        """Check and validate the fields in header part of a JWS object."""
        check_crit_header(header)
        validate_registry_header(self.header_registry, header)

        alg = self.get_alg(header["alg"])
        if alg.more_header_registry:
            validate_registry_header(alg.more_header_registry, header, check_more)

            if self.strict_check_header:
                allowed_registry = self.header_registry.copy()
                allowed_registry.update(alg.more_header_registry)
                check_supported_header(allowed_registry, header)
        elif self.strict_check_header:
            check_supported_header(self.header_registry, header)

    def get_alg(self, name: str) -> JWEAlgModel:
        """Get the allowed ("alg") algorithm instance of the given name.

        :param name: value of the ``alg``, e.g. ``ECDH-ES``, ``A128KW``
        """
        return self._get_algorithm("alg", name)

    def get_enc(self, name: str) -> JWEEncModel:
        """Get the allowed ("enc") algorithm instance of the given name.

        :param name: value of the ``enc``, e.g. ``A128CBC-HS256``, ``A128GCM``
        """
        return self._get_algorithm("enc", name)

    def get_zip(self, name: str) -> JWEZipModel:
        """Get the allowed ("zip") algorithm instance of the given name.

        :param name: value of the ``zip``, e.g. ``DEF``
        """
        return self._get_algorithm("zip", name)

    def _get_algorithm(self, location: t.Literal["alg", "enc", "zip"], name: str):
        registry: t.Dict[str, JWEAlgorithm] = self.algorithms[location]  # type: ignore
        if name not in registry:
            raise ValueError(f'Algorithm of "{name}" is not supported')

        if self.allowed:
            allowed = self.allowed
        else:
            allowed = self.recommended

        if name not in allowed:
            raise ValueError(f'Algorithm of "{name}" is not allowed')
        return registry[name]


default_registry = JWERegistry()



"""
    authlib.jose.rfc7518
    ~~~~~~~~~~~~~~~~~~~~

    Cryptographic Models for Cryptographic Models for Content
    Encryption per `Section 5`_.

    .. _`Section 5`: https://tools.ietf.org/html/rfc7518#section-5
"""
import typing as t
import hmac
import hashlib
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import GCM, CBC
from cryptography.hazmat.primitives.padding import PKCS7
from cryptography.exceptions import InvalidTag
from ..rfc7516.models import JWEEncModel
from ..errors import DecodeError
from .util import encode_int


class CBCHS2EncModel(JWEEncModel):
    # The IV used is a 128-bit value generated randomly or
    # pseudo-randomly for use in the cipher.
    iv_size = 128
    recommended = True

    def __init__(self, key_size: int, hash_type: int):
        self.name = f"A{key_size}CBC-HS{hash_type}"
        self.description = (
            f"AES_{key_size}_CBC_HMAC_SHA_{hash_type} authenticated encryption algorithm"
        )

        # key size in bit
        self.key_size = key_size
        # key size in byte
        self.key_len = key_size // 8

        self.cek_size = key_size * 2
        self.hash_alg = getattr(hashlib, f"sha{hash_type}")

    def _hmac(self, ciphertext: bytes, aad: bytes, iv: bytes, key: bytes) -> bytes:
        al = encode_int(len(aad) * 8, 64)
        msg = aad + iv + ciphertext + al
        d = hmac.new(key, msg, self.hash_alg).digest()
        return d[: self.key_len]

    def encrypt(self, plaintext: bytes, cek: bytes, iv: bytes, aad: bytes) -> t.Tuple[bytes, bytes]:
        """Key Encryption with AES_CBC_HMAC_SHA2."""
        hkey = cek[:self.key_len]
        ekey = cek[self.key_len:]

        pad = PKCS7(AES.block_size).padder()
        padded_data = pad.update(plaintext) + pad.finalize()

        cipher = Cipher(AES(ekey), CBC(iv), backend=default_backend())
        enc = cipher.encryptor()
        ciphertext = enc.update(padded_data) + enc.finalize()
        tag = self._hmac(ciphertext, aad, iv, hkey)
        return ciphertext, tag

    def decrypt(self, ciphertext: bytes, tag: bytes, cek: bytes, iv: bytes, aad: bytes) -> bytes:
        """Key Decryption with AES AES_CBC_HMAC_SHA2."""
        hkey = cek[:self.key_len]
        dkey = cek[self.key_len:]

        ctag = self._hmac(ciphertext, aad, iv, hkey)
        if not hmac.compare_digest(ctag, tag):
            raise DecodeError("tag does not match")

        cipher = Cipher(AES(dkey), CBC(iv), backend=default_backend())
        d = cipher.decryptor()
        data = d.update(ciphertext) + d.finalize()
        unpad = PKCS7(AES.block_size).unpadder()
        return unpad.update(data) + unpad.finalize()


class GCMEncModel(JWEEncModel):
    # Use of an IV of size 96 bits is REQUIRED with this algorithm.
    # https://tools.ietf.org/html/rfc7518#section-5.3
    iv_size = 96
    recommended = True

    def __init__(self, key_size: int):
        self.name = f"A{key_size}GCM"
        self.description = f"AES GCM using {key_size}-bit key"
        self.key_size = key_size
        self.cek_size = key_size

    def encrypt(self, plaintext: bytes, cek: bytes, iv: bytes, aad: bytes) -> t.Tuple[bytes, bytes]:
        """Key Encryption with AES GCM"""
        cipher = Cipher(AES(cek), GCM(iv), backend=default_backend())
        enc = cipher.encryptor()
        enc.authenticate_additional_data(aad)
        ciphertext = enc.update(plaintext) + enc.finalize()
        return ciphertext, enc.tag

    def decrypt(self, ciphertext: bytes, tag: bytes, cek: bytes, iv: bytes, aad: bytes) -> bytes:
        """Key Decryption with AES GCM"""
        cipher = Cipher(AES(cek), GCM(iv, tag), backend=default_backend())
        d = cipher.decryptor()
        d.authenticate_additional_data(aad)
        try:
            return d.update(ciphertext) + d.finalize()
        except InvalidTag as error:
            raise DecodeError(str(error))


JWE_ENC_MODELS: t.List[JWEEncModel] = [
    CBCHS2EncModel(128, 256),  # A128CBC-HS256
    CBCHS2EncModel(192, 384),  # A192CBC-HS384
    CBCHS2EncModel(256, 512),  # A256CBC-HS512
    GCMEncModel(128),  # A128GCM
    GCMEncModel(192),  # A192GCM
    GCMEncModel(256),  # A256GCM
]



import json
import typing as t
from joserfc.jwk import JWKRegistry
from unittest import TestCase
from pathlib import Path

BASE_PATH = Path(__file__).parent


def read_fixture(filename: str):
    with open((BASE_PATH / "fixtures" / filename).resolve()) as f:
        return json.load(f)


def load_key(filename: str, parameters=None):
    with open((BASE_PATH / "keys" / filename).resolve(), "rb") as f:
        content: bytes = f.read()

    if filename.endswith(".json"):
        data = json.loads(content)
        return JWKRegistry.import_key(data, parameters=parameters)

    kty = filename.split('-', 1)[0]
    return JWKRegistry.import_key(content, kty.upper(), parameters)


class TestFixture(TestCase):
    @classmethod
    def load_fixture(cls, filename: str):
        fixture_data = read_fixture(filename)

        for case_data in fixture_data['tests']:
            if "payload" not in case_data and "payload" in fixture_data:
                case_data["payload"] = fixture_data["payload"]
            cls.attach_case(case_data)

    @classmethod
    def attach_case(cls, data):
        runner = data.get("runner", "run_test")

        def method(self):
            getattr(self, runner)(data)

        case_name = data['name']
        name = f'test_{case_name}'
        method.__name__ = name
        method.__doc__ = f'Run fixture {data}'
        setattr(cls, name, method)

    def run_test(self, data: t.Dict[str, t.Any]):
        raise NotImplementedError()
